"""
Pandora API Client

This module contains the top level API client that is responsible for calling
the API and returing the results in model format. There is a base API client
that is useful for lower level programming such as calling methods that aren't
directly supported by the higher level API client.

The high level API client is what most clients should use and provides API
calls that map directly to the Pandora API and return model objects with
mappings from the raw JSON structures to Python objects.

For simplicity use a client builder from pandora.clientbuilder to create an
instance of a client.
"""
from . import errors
from .util import deprecated


class BaseAPIClient(object):
    """Base Pandora API Client

    The base API client has lower level methods that are composed together to
    provide higher level functionality.
    """

    LOW_AUDIO_QUALITY = "lowQuality"
    MED_AUDIO_QUALITY = "mediumQuality"
    HIGH_AUDIO_QUALITY = "highQuality"

    ALL_QUALITIES = [LOW_AUDIO_QUALITY, MED_AUDIO_QUALITY, HIGH_AUDIO_QUALITY]

    def __init__(self, transport, partner_user, partner_password, device,
                 default_audio_quality=MED_AUDIO_QUALITY):
        self.transport = transport
        self.partner_user = partner_user
        self.partner_password = partner_password
        self.device = device
        self.default_audio_quality = default_audio_quality
        self.username = None
        self.password = None

    @classmethod
    @deprecated("1.3", "2.0",
                "Replaced by clientbuilder.SettingsDictBuilder")
    def from_settings_dict(cls, settings):
        from .clientbuilder import SettingsDictBuilder
        return SettingsDictBuilder(settings).build()

    @classmethod
    @deprecated("1.3", "2.0",
                "Replaced by clientbuilder.PydoraConfigFileBuilder")
    def from_config_file(cls, path, authenticate=True):
        from .clientbuilder import PydoraConfigFileBuilder
        return PydoraConfigFileBuilder(path, authenticate).build()

    def _partner_login(self):
        partner = self.transport("auth.partnerLogin",
                                 username=self.partner_user,
                                 password=self.partner_password,
                                 deviceModel=self.device,
                                 version=self.transport.API_VERSION)

        self.transport.set_partner(partner)

        return partner

    def login(self, username, password):
        self.username = username
        self.password = password
        return self._authenticate()

    def _authenticate(self):
        self._partner_login()

        try:
            user = self.transport("auth.userLogin",
                                  loginType="user",
                                  username=self.username,
                                  password=self.password,
                                  includePandoraOneInfo=True,
                                  includeSubscriptionExpiration=True,
                                  returnCapped=True,
                                  includeAdAttributes=True,
                                  includeAdvertiserAttributes=True,
                                  xplatformAdCapable=True)
        except errors.InvalidPartnerLogin:
            raise errors.InvalidUserLogin()

        self.transport.set_user(user)

        return user

    @classmethod
    def get_qualities(cls, start_at, return_all_if_invalid=True):
        try:
            idx = cls.ALL_QUALITIES.index(start_at)
            return cls.ALL_QUALITIES[:idx + 1]
        except ValueError:
            if return_all_if_invalid:
                return cls.ALL_QUALITIES[:]
            else:
                return []

    def __call__(self, method, **kwargs):
        try:
            return self.transport(method, **kwargs)
        except errors.InvalidAuthToken:
            self._authenticate()
            return self.transport(method, **kwargs)


class APIClient(BaseAPIClient):
    """High Level Pandora API Client

    The high level API client implements the entire functional API for Pandora.
    This is what clients should actually use.
    """

    def get_station_list(self):
        from .models.pandora import StationList

        return StationList.from_json(self,
                                     self("user.getStationList",
                                          includeStationArtUrl=True))

    def get_station_list_checksum(self):
        return self("user.getStationListChecksum")["checksum"]

    def get_playlist(self, station_token):
        from .models.pandora import Playlist

        playlist = Playlist.from_json(self,
                                      self("station.getPlaylist",
                                           stationToken=station_token,
                                           includeTrackLength=True,
                                           xplatformAdCapable=True,
                                           audioAdPodCapable=True))

        for i, track in enumerate(playlist):
            if track.is_ad:
                track = self.get_ad_item(station_token, track.ad_token)
                playlist[i] = track

        return playlist

    def get_bookmarks(self):
        from .models.pandora import BookmarkList

        return BookmarkList.from_json(self,
                                      self("user.getBookmarks"))

    def get_station(self, station_token):
        from .models.pandora import Station

        return Station.from_json(self,
                                 self("station.getStation",
                                      stationToken=station_token,
                                      includeExtendedAttributes=True))

    def add_artist_bookmark(self, track_token):
        return self("bookmark.addArtistBookmark",
                    trackToken=track_token)

    def add_song_bookmark(self, track_token):
        return self("bookmark.addSongBookmark",
                    trackToken=track_token)

    def delete_song_bookmark(self, bookmark_token):
        return self("bookmark.deleteSongBookmark",
                    bookmarkToken=bookmark_token)

    def delete_artist_bookmark(self, bookmark_token):
        return self("bookmark.deleteArtistBookmark",
                    bookmarkToken=bookmark_token)

    def search(self, search_text,
               include_near_matches=False,
               include_genre_stations=False):
        from .models.pandora import SearchResult

        return SearchResult.from_json(
            self,
            self("music.search",
                 searchText=search_text,
                 includeNearMatches=include_near_matches,
                 includeGenreStations=include_genre_stations)
        )

    def add_feedback(self, track_token, positive):
        return self("station.addFeedback",
                    trackToken=track_token,
                    isPositive=positive)

    def add_music(self, music_token, station_token):
        return self("station.addMusic",
                    musicToken=music_token,
                    stationToken=station_token)

    def create_station(self, search_token=None, artist_token=None,
                       track_token=None):
        kwargs = {}

        if search_token:
            kwargs = {"musicToken": search_token}
        elif artist_token:
            kwargs = {"trackToken": artist_token, "musicType": "artist"}
        elif track_token:
            kwargs = {"trackToken": track_token, "musicType": "song"}
        else:
            raise KeyError("Must pass a type of token")

        return self("station.createStation", **kwargs)

    def delete_feedback(self, feedback_id):
        return self("station.deleteFeedback",
                    feedbackId=feedback_id)

    def delete_music(self, seed_id):
        return self("station.deleteMusic",
                    seedId=seed_id)

    def delete_station(self, station_token):
        return self("station.deleteStation",
                    stationToken=station_token)

    def get_genre_stations(self):
        from .models.pandora import GenreStationList

        genres = self("station.getGenreStations")

        genre_stations = GenreStationList.from_json(self, genres)
        genre_stations.checksum = self.get_genre_stations_checksum()
        return genre_stations

    def get_genre_stations_checksum(self):
        return self("station.getGenreStationsChecksum")["checksum"]

    def rename_station(self, station_token, name):
        return self("station.renameStation",
                    stationToken=station_token,
                    stationName=name)

    def explain_track(self, track_token):
        return self("track.explainTrack",
                    trackToken=track_token)

    def set_quick_mix(self, *args):
        return self("user.setQuickMix",
                    quickMixStationIds=args)

    def sleep_song(self, track_token):
        return self("user.sleepSong",
                    trackToken=track_token)

    def share_station(self, station_id, station_token, *emails):
        return self("station.shareStation",
                    stationId=station_id,
                    stationToken=station_token,
                    emails=emails)

    def transform_shared_station(self, station_token):
        return self("station.transformSharedStation",
                    stationToken=station_token)

    def share_music(self, music_token, *emails):
        return self("music.shareMusic",
                    musicToken=music_token,
                    email=emails[0])

    def get_ad_item(self, station_id, ad_token):
        from .models.pandora import AdItem

        if not station_id:
            raise errors.ParameterMissing("The 'station_id' param must be "
                                          "defined, got: '{}'"
                                          .format(station_id))

        ad_item = AdItem.from_json(self, self.get_ad_metadata(ad_token))
        ad_item.station_id = station_id
        ad_item.ad_token = ad_token
        return ad_item

    def get_ad_metadata(self, ad_token):
        return self("ad.getAdMetadata",
                    adToken=ad_token,
                    returnAdTrackingTokens=True,
                    supportAudioAds=True)

    def register_ad(self, station_id, tokens):
        return self("ad.registerAd",
                    stationId=station_id,
                    adTrackingTokens=tokens)



#!/usr/bin/env python
"""
Sample Barebones Pandora Player

This is a very simple Pandora player that streams music from Pandora. It
requires mpg123 to function. No songs are downloaded, they are streamed
directly from Pandora's servers.
"""
from __future__ import print_function

import os
import sys
from pandora import APIClient, clientbuilder

from .mpg123 import Player
from .utils import Colors, Screen


class PlayerCallbacks(object):
    """Interface for Player Callbacks

    This class simply exists to document the interface for callback
    implementers implementers need not extend this class.
    """

    def play(self, song):
        """Called once when a song starts playing
        """
        pass

    def pre_poll(self):
        """Called before polling for process status
        """
        pass

    def post_poll(self):
        """Called after polling for process status
        """
        pass

    def input(self, value, song):
        """Called after user input during song playback
        """
        pass


class PlayerApp(object):

    CMD_MAP = {
        "n": ("play next song", "skip_song"),
        "p": ("pause/resume song", "pause_song"),
        "s": ("stop playing station", "stop_station"),
        "d": ("dislike song", "dislike_song"),
        "u": ("like song", "like_song"),
        "b": ("bookmark song", "bookmark_song"),
        "a": ("bookmark artist", "bookmark_artist"),
        "S": ("sleep song for 30 days", "sleep_song"),
        "Q": ("quit player", "quit"),
        "?": ("display this help", "help"),
    }

    def __init__(self):
        self.client = None
        self.player = Player(self, sys.stdin)

    def get_client(self):
        cfg_file = os.environ.get("PYDORA_CFG", "")
        builder = clientbuilder.PydoraConfigFileBuilder(cfg_file)
        if builder.file_exists:
            return builder.build()

        builder = clientbuilder.PianobarConfigFileBuilder()
        if builder.file_exists:
            return builder.build()

        if not self.client:
            Screen.print_error("No valid config found")
            sys.exit(1)

    def station_selection_menu(self):
        """Format a station menu and make the user select a station
        """
        Screen.clear()

        for i, s in enumerate(self.stations):
            i = "{:>3}".format(i)
            print("{}: {}".format(Colors.yellow(i), s.name))

        return self.stations[Screen.get_integer("Station: ")]

    def play(self, song):
        """Play callback
        """
        if song.is_ad:
            print("{} ".format(Colors.cyan("Advertisement")))
        else:
            print("{} by {}".format(Colors.cyan(song.song_name),
                  Colors.yellow(song.artist_name)))

    def skip_song(self, song):
        if song.is_ad:
            Screen.print_error("Cannot skip advertisements")
        else:
            self.player.stop()

    def pause_song(self, song):
        self.player.pause()

    def stop_station(self, song):
        self.player.end_station()

    def dislike_song(self, song):
        try:
            if song.thumbs_down():
                Screen.print_success("Track disliked")
                self.player.stop()
            else:
                Screen.print_error("Failed to dislike track")
        except NotImplementedError:
            Screen.print_error("Cannot dislike this type of track")

    def like_song(self, song):
        try:
            if song.thumbs_up():
                Screen.print_success("Track liked")
            else:
                Screen.print_error("Failed to like track")
        except NotImplementedError:
            Screen.print_error("Cannot like this type of track")

    def bookmark_song(self, song):
        try:
            if song.bookmark_song():
                Screen.print_success("Bookmarked song")
            else:
                Screen.print_error("Failed to bookmark song")
        except NotImplementedError:
            Screen.print_error("Cannot bookmark this type of track")

    def bookmark_artist(self, song):
        try:
            if song.bookmark_artist():
                Screen.print_success("Bookmarked artist")
            else:
                Screen.print_error("Failed to bookmark artis")
        except NotImplementedError:
            Screen.print_error("Cannot bookmark artist for this type of track")

    def sleep_song(self, song):
        try:
            if song.sleep():
                Screen.print_success("Song will not be played for 30 days")
                self.player.stop()
            else:
                Screen.print_error("Failed to sleep song")
        except NotImplementedError:
            Screen.print_error("Cannot sleep this type of track")

    def quit(self, song):
        self.player.end_station()
        sys.exit(0)

    def help(self, song):
        print("")
        print("\n".join([
            "\t{} - {}".format(k, v[0]) for k, v in self.CMD_MAP.items()
        ]))
        print("")

    def input(self, input, song):
        """Input callback, handles key presses
        """
        try:
            cmd = getattr(self, self.CMD_MAP[input][1])
        except (IndexError, KeyError):
            return Screen.print_error("Invalid command!")

        cmd(song)

    def pre_poll(self):
        Screen.set_echo(False)

    def post_poll(self):
        Screen.set_echo(True)

    def run(self):
        self.client = self.get_client()
        self.stations = self.client.get_station_list()

        while True:
            try:
                station = self.station_selection_menu()
                self.player.play_station(station)
            except KeyboardInterrupt:
                sys.exit(0)


def main():
    PlayerApp().run()



"""
Pandora API Transport

This module contains the very low level transport agent for the Pandora API.
The transport is concerned with the details of a raw HTTP call to the Pandora
API along with the request and response encryption by way of an Encyrpytor
object. The result from a transport is a JSON object for the API or an
exception.

API consumers should use one of the API clients in the pandora.client package.
"""
import random
import time
import json
import base64
import requests
from requests.adapters import HTTPAdapter
from Crypto.Cipher import Blowfish

from .errors import PandoraException


DEFAULT_API_HOST = "tuner.pandora.com/services/json/"


def retries(max_tries, exceptions=(Exception,)):
    """Function decorator implementing retrying logic.

    exceptions: A tuple of exception classes; default (Exception,)

    The decorator will call the function up to max_tries times if it raises
    an exception.

    By default it catches instances of the Exception class and subclasses.
    This will recover after all but the most fatal errors. You may specify a
    custom tuple of exception classes with the 'exceptions' argument; the
    function will only be retried if it raises one of the specified
    exceptions.
    """
    def decorator(func):
        def function(*args, **kwargs):

            retries_left = max_tries
            while retries_left > 0:
                try:
                    retries_left -= 1
                    return func(*args, **kwargs)

                except exceptions as e:
                    # Don't retry for PandoraExceptions - unlikely that result
                    # will change for same set of input parameters.
                    if isinstance(e, PandoraException):
                        raise
                    if retries_left > 0:
                        time.sleep(delay_exponential(
                            0.5, 2, max_tries - retries_left))
                    else:
                        raise

        return function

    return decorator


def delay_exponential(base, growth_factor, attempts):
    """Calculate time to sleep based on exponential function.
    The format is::

        base * growth_factor ^ (attempts - 1)

    If ``base`` is set to 'rand' then a random number between
    0 and 1 will be used as the base.
    Base must be greater than 0, otherwise a ValueError will be
    raised.
    """
    if base == 'rand':
        base = random.random()
    elif base <= 0:
        raise ValueError("The 'base' param must be greater than 0, "
                         "got: {}".format(base))
    time_to_sleep = base * (growth_factor ** (attempts - 1))
    return time_to_sleep


class RetryingSession(requests.Session):
    """Requests Session With Retry Support

    This Requests session uses an HTTPAdapter that retries on connection
    failure three times. The Pandora API is fairly aggressive about closing
    connections on clients and the default session doesn't retry.
    """

    def __init__(self):
        super(RetryingSession, self).__init__()
        self.mount('https://', HTTPAdapter(max_retries=3))
        self.mount('http://', HTTPAdapter(max_retries=3))


class APITransport(object):
    """Pandora API Transport

    The transport is responsible for speaking the low-level protocol required
    by the Pandora API. It knows about encryption, TLS and the other API
    details. Once setup the transport acts like a callable.
    """

    API_VERSION = "5"

    REQUIRE_RESET = ("auth.partnerLogin", )
    NO_ENCRYPT = ("auth.partnerLogin", )
    REQUIRE_TLS = ("auth.partnerLogin", "auth.userLogin",
                   "station.getPlaylist", "user.createUser")

    def __init__(self, cryptor, api_host=DEFAULT_API_HOST, proxy=None):
        self.cryptor = cryptor
        self.api_host = api_host
        self._http = RetryingSession()

        if proxy:
            self._http.proxies = {"http": proxy, "https": proxy}

        self.reset()

    def reset(self):
        self.partner_auth_token = None
        self.user_auth_token = None

        self.partner_id = None
        self.user_id = None

        self.start_time = None
        self.server_sync_time = None

    def set_partner(self, data):
        self.sync_time = data["syncTime"]
        self.partner_auth_token = data["partnerAuthToken"]
        self.partner_id = data["partnerId"]

    def set_user(self, data):
        self.user_id = data["userId"]
        self.user_auth_token = data["userAuthToken"]

    @property
    def auth_token(self):
        if self.user_auth_token:
            return self.user_auth_token

        if self.partner_auth_token:
            return self.partner_auth_token

        return None

    @property
    def sync_time(self):
        if not self.server_sync_time:
            return None

        return int(self.server_sync_time + (time.time() - self.start_time))

    def remove_empty_values(self, data):
        return dict((k, v) for k, v in data.items() if v is not None)

    @sync_time.setter
    def sync_time(self, sync_time):
        self.server_sync_time = self.cryptor.decrypt_sync_time(sync_time)

    def _start_request(self, method):
        if method in self.REQUIRE_RESET:
            self.reset()

        if not self.start_time:
            self.start_time = int(time.time())

    def _make_http_request(self, url, data, params):
        try:
            data = data.encode("utf-8")
        except AttributeError:
            pass

        params = self.remove_empty_values(params)

        r = self._http.post(url, data=data, params=params)
        r.raise_for_status()
        return r.content

    def test_url(self, url):
        return self._http.head(url).status_code == requests.codes.OK

    def _build_params(self, method):
        return {
            "method": method,
            "auth_token": self.auth_token,
            "partner_id": self.partner_id,
            "user_id": self.user_id,
        }

    def _build_url(self, method):
        return "{0}://{1}".format(
            "https" if method in self.REQUIRE_TLS else "http",
            self.api_host)

    def _build_data(self, method, data):
        data["userAuthToken"] = self.user_auth_token

        if not self.user_auth_token and self.partner_auth_token:
            data["partnerAuthToken"] = self.partner_auth_token

        data["syncTime"] = self.sync_time
        data = json.dumps(self.remove_empty_values(data))

        if method not in self.NO_ENCRYPT:
            data = self.cryptor.encrypt(data)

        return data

    def _parse_response(self, result):
        result = json.loads(result.decode("utf-8"))

        if result["stat"] == "ok":
            return result["result"] if "result" in result else None
        else:
            raise PandoraException.from_code(result["code"], result["message"])

    # TODO: This decorator is a temporary workaround for handling
    # SysCallErrors, see: https://github.com/shazow/urllib3/issues/367.
    # Should be removed once a fix is applied in urllib3.
    @retries(3)
    def __call__(self, method, **data):
        self._start_request(method)

        url = self._build_url(method)
        data = self._build_data(method, data)
        params = self._build_params(method)
        result = self._make_http_request(url, data, params)

        return self._parse_response(result)


class Encryptor(object):
    """Pandora Blowfish Encryptor

    The blowfish encryptor can encrypt and decrypt the relevant parts of the
    API request and response. It handles the formats that the API expects.
    """

    def __init__(self, in_key, out_key):
        self.bf_out = Blowfish.new(out_key, Blowfish.MODE_ECB)
        self.bf_in = Blowfish.new(in_key, Blowfish.MODE_ECB)

    @staticmethod
    def _decode_hex(data):
        return base64.b16decode(data.encode("ascii").upper())

    @staticmethod
    def _encode_hex(data):
        return base64.b16encode(data).lower()

    def decrypt(self, data):
        data = self.bf_out.decrypt(self._decode_hex(data))
        return json.loads(self.strip_padding(data))

    def decrypt_sync_time(self, data):
        return int(self.bf_in.decrypt(self._decode_hex(data))[4:-2])

    def add_padding(self, data):
        block_size = Blowfish.block_size
        pad_size = len(data) % block_size
        return data + (chr(pad_size) * (block_size - pad_size))

    def strip_padding(self, data):
        pad_size = int(data[-1])
        if not data[-pad_size:] == bytes((pad_size,)) * pad_size:
            raise ValueError('Invalid padding')
        return data[:-pad_size]

    def encrypt(self, data):
        return self._encode_hex(self.bf_out.encrypt(self.add_padding(data)))



import select

from .utils import iterate_forever, SilentPopen


class Player(object):
    """Remote control for an mpg123 process

    Starts and owns a handle to an mpg123 process then feeds commands to it to
    play pandora audio
    """

    def __init__(self, callbacks, control_channel):
        self._control_channel = control_channel
        self._control_fd = control_channel.fileno()
        self._callbacks = callbacks
        self._process = None
        self._ensure_started()

    def __del__(self):
        self._process.kill()

    def _ensure_started(self):
        """Ensure mpg123 is started
        """
        if self._process and self._process.poll() is None:
            return

        self._process = SilentPopen(
            ["mpg123", "-q", "-R", "--ignore-mime"])

        # Only output play status in the player stdout
        self._send_cmd("silence")

    def _send_cmd(self, cmd):
        """Write command to remote mpg123 process
        """
        self._process.stdin.write("{}\n".format(cmd).encode("utf-8"))
        self._process.stdin.flush()

    def stop(self):
        """Stop the currently playing song
        """
        self._send_cmd("stop")

    def pause(self):
        """Pause the player
        """
        self._send_cmd("pause")

    def _player_stopped(self, value):
        """Determine if player has stopped
        """
        return value.startswith(b"@P") and value.decode("utf-8")[3] == "0"

    def play(self, song):
        """Play a new song from a Pandora model

        Returns once the stream starts but does not shut down the remote mpg123
        process. Calls the input callback when the user has input.
        """
        self._callbacks.play(song)
        self._send_cmd("load {}".format(song.audio_url))

        while True:
            try:
                self._callbacks.pre_poll()
                self._ensure_started()

                readers, _, _ = select.select(
                    [self._control_channel, self._process.stdout], [], [], 1)

                for fd in readers:
                    value = fd.readline().strip()

                    if fd.fileno() == self._control_fd:
                        self._callbacks.input(value, song)
                    else:
                        if self._player_stopped(value):
                            return
            finally:
                self._callbacks.post_poll()

    def end_station(self):
        """Stop playing the station
        """
        raise StopIteration

    def play_station(self, station):
        """Play the station until something ends it

        This function will run forever until termintated by calling
        end_station.
        """
        for song in iterate_forever(station.get_playlist):
            try:
                self.play(song)
            except StopIteration:
                self.stop()
                return



from datetime import datetime
from collections import namedtuple


def with_metaclass(meta, *bases):
    return meta("NewBase", bases, {})


class Field(namedtuple("Field", ["field", "default", "formatter"])):

    def __new__(cls, field, default=None, formatter=None):
        return super(Field, cls).__new__(cls, field, default, formatter)


class ModelMetaClass(type):

    def __new__(cls, name, parents, dct):
        dct["_fields"] = fields = {}
        new_dct = dct.copy()

        for key, val in dct.items():
            if key.startswith("__"):
                continue

            if isinstance(val, Field):
                fields[key] = val
                del new_dct[key]

        return super(ModelMetaClass, cls).__new__(cls, name, parents, new_dct)


class PandoraModel(with_metaclass(ModelMetaClass, object)):

    @staticmethod
    def json_to_date(api_client, data):
        return datetime.utcfromtimestamp(data["time"] / 1000)

    @classmethod
    def from_json_list(cls, api_client, data):
        return [cls.from_json(api_client, item) for item in data]

    def __init__(self, api_client):
        self._api_client = api_client

        safe_types = (type(None), str, bytes, int, bool)

        for key, value in self._fields.items():
            default = value.default

            if not isinstance(default, safe_types):
                default = type(default)()

            setattr(self, key, default)

    @staticmethod
    def populate_fields(api_client, instance, data):
        for key, value in instance.__class__._fields.items():
            newval = data.get(value.field, value.default)

            if newval and value.formatter:
                newval = value.formatter(api_client, newval)

            setattr(instance, key, newval)

    @classmethod
    def from_json(cls, api_client, data):
        self = cls(api_client)
        PandoraModel.populate_fields(api_client, self, data)
        return self

    def _base_repr(self, and_also=None):
        items = [
            "=".join((key, repr(getattr(self, key))))
            for key in sorted(self._fields.keys())]

        if items:
            output = ", ".join(items)
        else:
            output = None

        if and_also:
            return "{}({}, {})".format(self.__class__.__name__,
                                       output, and_also)
        else:
            return "{}({})".format(self.__class__.__name__, output)

    def __repr__(self):
        return self._base_repr()


class PandoraListModel(PandoraModel, list):

    __list_key__ = None
    __list_model__ = None
    __index_key__ = None

    def __init__(self, *args, **kwargs):
        super(PandoraListModel, self).__init__(*args, **kwargs)
        self._index = {}

    @classmethod
    def from_json(cls, api_client, data):
        self = cls(api_client)
        PandoraModel.populate_fields(api_client, self, data)

        for item in data[cls.__list_key__]:
            model = cls.__list_model__.from_json(api_client, item)

            if self.__index_key__:
                value = getattr(model, self.__index_key__)
                self._index[value] = model

            self.append(model)

        return self

    def __getitem__(self, key):
        item = self._index.get(key, None)
        if item:
            return item
        else:
            return list.__getitem__(self, key)

    def __contains__(self, key):
        if key in self._index:
            return True
        else:
            return list.__contains__(self, key)

    def keys(self):
        return self._index.keys()

    def items(self):
        return self._index.items()

    def __repr__(self):
        return self._base_repr(and_also=list.__repr__(self))


class PandoraDictListModel(PandoraModel, dict):

    __dict_list_key__ = None
    __dict_key__ = None
    __list_key__ = None
    __list_model__ = None

    @classmethod
    def from_json(cls, api_client, data):
        self = cls(api_client)
        PandoraModel.populate_fields(api_client, self, data)

        if self.__dict_list_key__:
            data = data[self.__dict_list_key__]

        for item in data:
            key = item[self.__dict_key__]
            self[key] = []

            for part in item[self.__list_key__]:
                self[key].append(
                    cls.__list_model__.from_json(api_client, part))

        return self

    def __repr__(self):
        return self._base_repr(and_also=dict.__repr__(self))



from .. import BaseAPIClient
from . import with_metaclass, ModelMetaClass
from . import Field, PandoraModel, PandoraListModel, PandoraDictListModel
from ..errors import ParameterMissing


class Station(PandoraModel):

    can_add_music = Field("allowAddMusic")
    can_delete = Field("allowDelete")
    can_rename = Field("allowRename")
    is_shared = Field("isShared")
    is_quickmix = Field("isQuickMix")

    art_url = Field("artUrl")
    date_created = Field("dateCreated", formatter=PandoraModel.json_to_date)
    detail_url = Field("stationDetailUrl")
    id = Field("stationId")
    name = Field("stationName")
    sharing_url = Field("stationSharingUrl")
    token = Field("stationToken")

    genre = Field("genre", [])
    quickmix_stations = Field("quickMixStationIds", [])

    def get_playlist(self):
        return iter(self._api_client.get_playlist(self.token))


class GenreStation(PandoraModel):

    id = Field("stationId")
    name = Field("stationName")
    token = Field("stationToken")
    category = Field("categoryName")

    def get_playlist(self):
        raise NotImplementedError("Genre stations do not have playlists. "
                                  "Create a real station using the token.")


class StationList(PandoraListModel):

    checksum = Field("checksum")

    __index_key__ = "id"
    __list_key__ = "stations"
    __list_model__ = Station

    def has_changed(self):
        checksum = self._api_client.get_station_list_checksum()
        return checksum != self.checksum


class PlaylistModel(PandoraModel):

    @classmethod
    def from_json(cls, api_client, data):
        self = cls(api_client)

        for key, value in cls._fields.items():
            newval = data.get(value.field, value.default)

            if value.field == "audioUrl" and newval is None:
                newval = cls.get_audio_url(
                    data, api_client.default_audio_quality)

            if value.field == "bitrate" and newval is None:
                newval = cls.get_audio_bitrate(
                    data, api_client.default_audio_quality)

            if newval and value.formatter:
                newval = value.formatter(newval)

            setattr(self, key, newval)

        return self

    @classmethod
    def get_audio_field(cls, data, field, preferred_quality):
        """Get audio-related fields

        Try to find fields for the audio url for specified preferred quality
        level, or next-lowest available quality url otherwise.
        """
        audio_url = None
        url_map = data.get("audioUrlMap")

        # No audio url available (e.g. ad tokens)
        if not url_map:
            return None

        valid_audio_formats = [BaseAPIClient.HIGH_AUDIO_QUALITY,
                               BaseAPIClient.MED_AUDIO_QUALITY,
                               BaseAPIClient.LOW_AUDIO_QUALITY]

        # Only iterate over sublist, starting at preferred audio quality, or
        # from the beginning of the list if nothing is found. Ensures that the
        # bitrate used will always be the same or lower quality than was
        # specified to prevent audio from skipping for slow connections.
        if preferred_quality in valid_audio_formats:
            i = valid_audio_formats.index(preferred_quality)
            valid_audio_formats = valid_audio_formats[i:]

        for quality in valid_audio_formats:
            audio_url = url_map.get(quality)

            if audio_url:
                return audio_url[field]

        return audio_url[field] if audio_url else None

    @classmethod
    def get_audio_url(cls, data,
                      preferred_quality=BaseAPIClient.MED_AUDIO_QUALITY):
        """Get audio url

        Try to find audio url for specified preferred quality level, or
        next-lowest available quality url otherwise.
        """
        return cls.get_audio_field(data, "audioUrl", preferred_quality)

    @classmethod
    def get_audio_bitrate(cls, data,
                          preferred_quality=BaseAPIClient.MED_AUDIO_QUALITY):
        """Get audio bitrate

        Try to find bitrate of audio url for specified preferred quality level,
        or next-lowest available quality url otherwise.
        """
        return cls.get_audio_field(data, "bitrate", preferred_quality)

    def get_is_playable(self):
        if not self.audio_url:
            return False
        return self._api_client.transport.test_url(self.audio_url)

    def prepare_playback(self):
        """Prepare Track for Playback

        This method must be called by clients before beginning playback
        otherwise the track recieved may not be playable.
        """
        return self

    def thumbs_up(self):
        raise NotImplementedError

    def thumbs_down(self):
        raise NotImplementedError

    def bookmark_song(self):
        raise NotImplementedError

    def bookmark_artist(self):
        raise NotImplementedError

    def sleep(self):
        raise NotImplementedError


class PlaylistItem(PlaylistModel):

    artist_name = Field("artistName")
    album_name = Field("albumName")
    song_name = Field("songName")
    song_rating = Field("songRating")
    track_gain = Field("trackGain")
    track_length = Field("trackLength")
    track_token = Field("trackToken")
    audio_url = Field("audioUrl")
    bitrate = Field("bitrate")
    album_art_url = Field("albumArtUrl")
    allow_feedback = Field("allowFeedback")
    station_id = Field("stationId")

    ad_token = Field("adToken")

    album_detail_url = Field("albumDetailUrl")
    album_explore_url = Field("albumExplorerUrl")

    amazon_album_asin = Field("amazonAlbumAsin")
    amazon_album_digital_asin = Field("amazonAlbumDigitalAsin")
    amazon_album_url = Field("amazonAlbumUrl")
    amazon_song_digital_asin = Field("amazonSongDigitalAsin")

    artist_detail_url = Field("artistDetailUrl")
    artist_explore_url = Field("artistExplorerUrl")

    itunes_song_url = Field("itunesSongUrl")

    song_detail_url = Field("songDetailUrl")
    song_explore_url = Field("songExplorerUrl")

    @property
    def is_ad(self):
        return self.ad_token is not None

    def thumbs_up(self):
        return self._api_client.add_feedback(self.track_token, True)

    def thumbs_down(self):
        return self._api_client.add_feedback(self.track_token, False)

    def bookmark_song(self):
        return self._api_client.add_song_bookmark(self.track_token)

    def bookmark_artist(self):
        return self._api_client.add_artist_bookmark(self.track_token)

    def sleep(self):
        return self._api_client.sleep_song(self.track_token)


class AdItem(PlaylistModel):

    title = Field("title")
    company_name = Field("companyName")
    tracking_tokens = Field("adTrackingTokens")
    audio_url = Field("audioUrl")
    image_url = Field("imageUrl")
    click_through_url = Field("clickThroughUrl")
    station_id = None
    ad_token = None

    @property
    def is_ad(self):
        return True

    def register_ad(self, station_id=None):
        if not station_id:
            station_id = self.station_id
        if self.tracking_tokens:
            self._api_client.register_ad(station_id, self.tracking_tokens)
        else:
            raise ParameterMissing('No ad tracking tokens provided for '
                                   'registration.')

    def prepare_playback(self):
        try:
            self.register_ad(self.station_id)
        except ParameterMissing as e:
            if self.tracking_tokens:
                raise e
        return super(AdItem, self).prepare_playback()


class Playlist(PandoraListModel):

    __list_key__ = "items"
    __list_model__ = PlaylistItem


class Bookmark(PandoraModel):

    music_token = Field("musicToken")
    artist_name = Field("artistName")
    art_url = Field("artUrl")
    bookmark_token = Field("bookmarkToken")
    date_created = Field("dateCreated", formatter=PandoraModel.json_to_date)

    # song only
    sample_url = Field("sampleUrl")
    sample_gain = Field("sampleGain")
    album_name = Field("albumName")
    song_name = Field("songName")

    @property
    def is_song_bookmark(self):
        return self.song_name is not None

    def delete(self):
        if self.is_song_bookmark:
            self._api_client.delete_song_bookmark(self.bookmark_token)
        else:
            self._api_client.delete_artist_bookmark(self.bookmark_token)


class BookmarkList(PandoraModel):

    songs = Field("songs", formatter=Bookmark.from_json_list)
    artists = Field("artists", formatter=Bookmark.from_json_list)


class SearchResultItem(PandoraModel):

    score = Field("score")
    token = Field("musicToken")

    @property
    def is_song(self):
        return isinstance(self, SongSearchResultItem)

    @property
    def is_artist(self):
        return isinstance(self, ArtistSearchResultItem) and \
               self.token.startswith("R")

    @property
    def is_composer(self):
        return isinstance(self, ArtistSearchResultItem) and \
               self.token.startswith("C")

    @property
    def is_genre_station(self):
        return isinstance(self, GenreStationSearchResultItem)

    def create_station(self):
        raise NotImplementedError

    @classmethod
    def from_json(cls, api_client, data):
        if data["musicToken"].startswith("S"):
            return SongSearchResultItem.from_json(api_client, data)

        elif data["musicToken"].startswith(("R", "C")):
            return ArtistSearchResultItem.from_json(api_client, data)

        elif data["musicToken"].startswith("G"):
            return GenreStationSearchResultItem.from_json(api_client, data)
        else:
            raise NotImplementedError("Unknown result token type '{}'"
                                      .format(data["musicToken"]))


class ArtistSearchResultItem(SearchResultItem):

    score = Field("score")
    token = Field("musicToken")
    artist = Field("artistName")
    likely_match = Field("likelyMatch", default=False)

    def create_station(self):
        self._api_client.create_station(artist_token=self.token)

    @classmethod
    def from_json(cls, api_client, data):
        return super(SearchResultItem, cls).from_json(api_client, data)


class SongSearchResultItem(SearchResultItem):

    score = Field("score")
    token = Field("musicToken")
    artist = Field("artistName")
    song_name = Field("songName")

    def create_station(self):
        self._api_client.create_station(track_token=self.token)

    @classmethod
    def from_json(cls, api_client, data):
        return super(SearchResultItem, cls).from_json(api_client, data)


class GenreStationSearchResultItem(SearchResultItem):

    score = Field("score")
    token = Field("musicToken")
    station_name = Field("stationName")

    def create_station(self):
        self._api_client.create_station(search_token=self.token)

    @classmethod
    def from_json(cls, api_client, data):
        return super(SearchResultItem, cls).from_json(api_client, data)


class SearchResult(PandoraModel):

    nearest_matches_available = Field("nearMatchesAvailable")
    explanation = Field("explanation")
    songs = Field("songs", formatter=SongSearchResultItem.from_json_list)
    artists = Field("artists", formatter=ArtistSearchResultItem.from_json_list)
    genre_stations = Field(
        "genreStations",
        formatter=GenreStationSearchResultItem.from_json_list)


class GenreStationList(PandoraDictListModel):

    checksum = Field("checksum")

    __dict_list_key__ = "categories"
    __dict_key__ = "categoryName"
    __list_key__ = "stations"
    __list_model__ = GenreStation

    def has_changed(self):
        checksum = self._api_client.get_station_list_checksum()
        return checksum != self.checksum



"""
Python 2 Compatibility Layer

This module exists to work around compatibility issues between Python 2 and
Python 3. The main code-base will use Python 3 idioms and this module will
patch Python 2 code to support those changes. When Python 2 support is
dropped this module can be removed and imports can be updated.
"""

try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import SafeConfigParser

    class ConfigParser(SafeConfigParser):

        def read_file(self, fp):
            return self.readfp(fp)


# Only used in tests
try:
    from unittest.mock import Mock, MagicMock, call, patch
except ImportError:
    try:
        from mock import Mock, MagicMock, call, patch
    except ImportError:
        pass



"""
Pandora API Client Builders

This module provides a set of builder classes that can turn various
configuration formats into a fully built APIClient.
"""
import os.path

from .py2compat import ConfigParser
from . import Encryptor, APITransport, DEFAULT_API_HOST, APIClient


class TranslatingDict(dict):
    """Abstract Key/Value Translating Dictionary

    Dictionary that translates keys using a static map of old key to new key
    and values using a map of key to value translating function. The value
    translating function will be called with the key and value read and is
    expected to return a translated value for storage. Keys and values not
    matched for translation are stored as provided.

    This otherwise behaves as a standard dictionary.

    Subclasses must provide KEY_TRANSLATIONS and VALUE_TRANSLATIONS even if
    they are just empty dictionaries.
    """

    KEY_TRANSLATIONS = None
    VALUE_TRANSLATIONS = None

    def __init__(self, initial=None):
        super(TranslatingDict, self).__init__()

        assert self.KEY_TRANSLATIONS is not None
        assert self.VALUE_TRANSLATIONS is not None

        if not initial:
            return

        if hasattr(initial, "items"):
            values = initial.items()
        else:
            values = initial

        for key, value in values:
            self.put(key, value)

    def was_translated(self, from_key, to_key):
        pass

    def translate_key(self, key):
        key = key.strip().upper()
        to_key = self.KEY_TRANSLATIONS.get(key, None)

        if to_key:
            self.was_translated(key, to_key)
            return to_key
        else:
            return key

    def translate_value(self, key, value):
        if hasattr(value, "strip"):
            value = value.strip()

        return self.VALUE_TRANSLATIONS.get(key, lambda v: v)(value)

    def put(self, key, value):
        self[key] = value

    def __setitem__(self, key, value):
        key = self.translate_key(key)
        super(TranslatingDict, self).__setitem__(
            key, self.translate_value(key, value))


class APIClientBuilder(object):
    """Abstract API Client Builder

    Provides the basic functions for building an API client. Expects a
    dictionary of standard configuration options.

    Required values:
    * DECRYPTION_KEY - Pandora API decryption key
    * ENCRYPTION_KEY - Pandora API encryption key
    * PARTNER_USER - Pandora API partner username
    * PARTNER_PASSWORD - Pandora API partner password
    * DEVICE - Pandora API device type identifier

    Optional values:
    * API_HOST - API hostname and path to API
    * PROXY - HTTP/HTTPS proxy hostname
    * AUDIO_QUALITY - A supported audio quality (see APIClient)
    """

    DEFAULT_CLIENT_CLASS = APIClient

    def __init__(self, client_class=None):
        self.client_class = client_class or self.DEFAULT_CLIENT_CLASS

    def build_from_settings_dict(self, settings):
        enc = Encryptor(settings["DECRYPTION_KEY"],
                        settings["ENCRYPTION_KEY"])

        trans = APITransport(enc,
                             settings.get("API_HOST", DEFAULT_API_HOST),
                             settings.get("PROXY", None))

        quality = settings.get("AUDIO_QUALITY",
                               self.client_class.MED_AUDIO_QUALITY)

        return self.client_class(trans, settings["PARTNER_USER"],
                                 settings["PARTNER_PASSWORD"],
                                 settings["DEVICE"], quality)


class SettingsDict(TranslatingDict):
    """Settings Translating Dictionary

    Maps old setting keys to new ones. Should be removed when ready to break
    backwards compatibility.
    """

    KEY_TRANSLATIONS = {
        "USERNAME": "PARTNER_USER",
        "PASSWORD": "PARTNER_PASSWORD",
        "DEFAULT_AUDIO_QUALITY": "AUDIO_QUALITY",
    }

    VALUE_TRANSLATIONS = {}

    def was_translated(self, from_key, to_key):
        # TODO: Notify of deprecations
        pass


class SettingsDictBuilder(APIClientBuilder):
    """Settings Dictionary Client Builder

    Builds an API client based on a translated settings dictionary.
    """

    def __init__(self, settings, **kwargs):
        self.settings = settings
        super(SettingsDictBuilder, self).__init__(**kwargs)

    def build(self):
        settings = SettingsDict(self.settings)
        return self.build_from_settings_dict(settings)


class FileBasedClientBuilder(APIClientBuilder):
    """Abstract File Based Client Builder

    Provides base functionality for client builders that load their settings
    from files.
    """

    DEFAULT_CONFIG_FILE = ""

    def __init__(self, path=None, authenticate=True, **kwargs):
        self.path = path or self.DEFAULT_CONFIG_FILE
        self.authenticate = authenticate
        super(FileBasedClientBuilder, self).__init__(**kwargs)

    @property
    def file_exists(self):
        return os.path.exists(self._path)

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, path):
        self._path = os.path.expanduser(path)

    def build(self):
        if not self.file_exists:
            raise IOError("File not found: {}".format(self.path))

        config = self.parse_config()
        client = self.build_from_settings_dict(config)

        if self.authenticate:
            client.login(config["USER"]["USERNAME"],
                         config["USER"]["PASSWORD"])

        return client


class PydoraConfigFileBuilder(FileBasedClientBuilder):
    """Pydora Config Format Client Builder

    Builds API client for original pydora configuration format.
    """

    DEFAULT_CONFIG_FILE = "~/.pydora.cfg"

    @staticmethod
    def cfg_to_dict(cfg, key, kind=SettingsDict):
        return kind((k.strip().upper(), v.strip())
                    for k, v in cfg.items(key, raw=True))

    def parse_config(self):
        cfg = ConfigParser()

        with open(self.path) as fp:
            cfg.read_file(fp)

        settings = PydoraConfigFileBuilder.cfg_to_dict(cfg, "api")
        settings["user"] = PydoraConfigFileBuilder.cfg_to_dict(
            cfg, "user", dict)

        return settings


class PianobarSettingsDict(TranslatingDict):
    """Pianobar Translating Dictionary
    """

    KEY_TRANSLATIONS = {
        "DECRYPT_PASSWORD": "DECRYPTION_KEY",
        "ENCRYPT_PASSWORD": "ENCRYPTION_KEY",
        "RPC_HOST": "API_HOST",
        "CONTROL_PROXY": "PROXY",
    }

    VALUE_TRANSLATIONS = {
        "API_HOST": lambda v: "{}/services/json/".format(v),
        "AUDIO_QUALITY": lambda v: "{}Quality".format(v),
    }


class PianobarConfigFileBuilder(FileBasedClientBuilder):
    """Pianobar Config File Client Builder

    Builds an API client from a Pianobar config file.
    """

    DEFAULT_CONFIG_FILE = "~/.config/pianobar/config"

    def parse_config(self):
        settings = PianobarSettingsDict()

        with open(self.path, "r") as fp:
            for line in fp.readlines():
                line = line.strip()

                if line and not line.startswith("#"):
                    settings.put(*line.split("=", 1))

        settings["USER"] = {
            "USERNAME": settings.pop("USER"),
            "PASSWORD": settings.pop("PASSWORD"),
        }

        return settings



import os
import re
import sys
import requests

from pandora.client import APIClient
from pandora.py2compat import ConfigParser
from pandora.clientbuilder import PydoraConfigFileBuilder

from .utils import Screen, Colors


class umask(object):
    """Set/Restore Umask Context Manager
    """

    def __init__(self, umask):
        self.umask = umask
        self.old_umask = None

    def __enter__(self):
        self.old_umask = os.umask(self.umask)

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.umask(self.old_umask)


class PandoraKeysConfigParser(object):
    """Parser for Pandora Keys Source Page

    This is an extremely naive restructured text parser designed only to parse
    the pandora API docs keys source file.
    """

    KEYS_URL = ("http://6xq.net/git/lars/pandora-apidoc.git/"
                "plain/json/partners.rst")

    FIELD_RE = re.compile(
            ":(?P<key>[^:]+): (?:`{2})?(?P<value>[^`\n]+)(?:`{2})?$")

    def _fixup_key(self, key):
        key = key.lower()

        if key.startswith("dec") and "password" in key:
            return "decryption_key"
        elif key.startswith("encrypt"):
            return "encryption_key"
        elif key == "deviceid":
            return "device"
        else:
            return key

    def _format_api_host(self, host):
        return "{}/services/json/".format(host)

    def _clean_device_name(self, name):
        return re.sub("[^a-z]+", "_", name, flags=re.I)

    def _fetch_config(self):
        return requests.get(self.KEYS_URL).text.split("\n")

    def _match_key(self, line):
        key_match = self.FIELD_RE.match(line)
        if key_match:
            match = key_match.groupdict()
            match["key"] = self._fixup_key(match["key"])
            return match
        else:
            return None

    def _is_host_terminator(self, line):
        return line.startswith("--")

    def _is_device_terminator(self, line):
        return line.startswith("^^")

    def load(self):
        buffer = []
        current_partner = {}
        api_host = None
        partners = {}

        for line in self._fetch_config():
            key_match = self._match_key(line)
            if key_match:
                current_partner[key_match["key"]] = key_match["value"]
            elif self._is_host_terminator(line):
                api_host = buffer.pop()
            elif self._is_device_terminator(line):
                key = self._clean_device_name(buffer.pop())
                current_partner = partners[key] = {
                        "api_host": self._format_api_host(api_host)
                        }

            buffer.append(line.strip().lower())

        return partners


class Configurator(object):
    """Interactive Configuration Builder

    Allows a user to configure pydora interactively. Ultimately writes the
    pydora config file.
    """

    def __init__(self):
        self.builder = PydoraConfigFileBuilder()

        self.cfg = ConfigParser()
        self.cfg.add_section("user")
        self.cfg.add_section("api")

    def fail(self, message):
        print(Screen.print_error(message))
        sys.exit(1)

    def finished(self, message):
        Screen.print_success(message)
        sys.exit(0)

    def print_message(self, message):
        print(Colors.cyan(message))

    def get_partner_config(self):
        try:
            return PandoraKeysConfigParser().load()["ios"]
        except:
            self.fail("Error loading config file. Unable to continue.")

    def get_value(self, section, key, prompt):
        self.cfg.set(section, key, Screen.get_string(prompt))

    def get_password(self, section, key, prompt):
        self.cfg.set(section, key, Screen.get_password(prompt))

    def set_static_value(self, section, key, value):
        self.cfg.set(section, key, value)

    def add_partner_config(self, config):
        for key, value in config.items():
            self.cfg.set("api", key, value)

    def write_config(self):
        with umask(0o077), open(self.builder.path, "w") as fp:
            self.cfg.write(fp)

    def configure(self):
        if self.builder.file_exists:
            self.finished("You already have a pydora config!")

        self.print_message("Welcome to Pydora, let's configure a few things")
        self.add_partner_config(self.get_partner_config())
        self.get_value("user", "username", "Pandora Username: ")
        self.get_password("user", "password", "Pandora Password: ")
        self.set_static_value("api", "default_audio_quality",
                              APIClient.HIGH_AUDIO_QUALITY)

        self.write_config()


def main():
    Configurator().configure()



"""
Utility Functions

Functions that don't have a home anywhere else.
"""

import warnings
from functools import wraps


def warn_deprecated(in_version, remove_version, what, message):
    """Warn that something is deprecated
    """
    msg = ("{} is deprecated as of version {}"
           " and will be removed in version {}. {}")

    warnings.warn(
        msg.format(what, in_version, remove_version, message),
        DeprecationWarning)


def deprecated(in_version, remove_version, message):
    """Deprecated function decorator

    Decorator to warn that a function is deprecated and what version it will be
    removed in.
    """
    def wrapper(f):
        @wraps(f)
        def inner_wrapper(self, *args, **kwargs):
            warn_deprecated(in_version, remove_version, f.func_name, message)
            return f(self, *args, **kwargs)
        return inner_wrapper
    return wrapper



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name="pydora",
    version="1.7.3",
    description="Python wrapper for Pandora API",
    long_description=open("README.rst", "r").read(),
    author="Mike Crute",
    author_email="mcrute@gmail.com",
    url="https://github.com/mcrute/pydora",
    test_suite="tests.discover_suite",
    packages=find_packages(exclude=["tests", "tests.*"]),
    setup_requires=[
        "py_release_tools",
        "wheel",
    ],
    install_requires=[
        "pycrypto>=2.6.1",
        "requests>=2",
    ],
    entry_points={
        "console_scripts": [
            "pydora = pydora.player:main",
            "pydora-configure = pydora.configure:main",
        ],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Multimedia :: Sound/Audio :: Players",
    ]
)



from __future__ import print_function

import os
import sys
import termios
import getpass
import subprocess
from pandora.models.pandora import AdItem
from pandora import errors


def input(prompt):
    try:
        return raw_input(prompt)
    except NameError:
        import builtins
        return builtins.input(prompt)


class Colors(object):

    def __wrap_with(code):
        @staticmethod
        def inner(text, bold=False):
            c = code
            if bold:
                c = "1;{}".format(c)
            return "\033[{}m{}\033[0m".format(c, text)
        return inner

    red = __wrap_with("31")
    green = __wrap_with("32")
    yellow = __wrap_with("33")
    blue = __wrap_with("34")
    magenta = __wrap_with("35")
    cyan = __wrap_with("36")
    white = __wrap_with("37")


class Screen(object):

    @staticmethod
    def set_echo(enabled):
        fd = sys.stdin.fileno()
        if not os.isatty(fd):
            return

        (iflag, oflag, cflag,
         lflag, ispeed, ospeed, cc) = termios.tcgetattr(fd)

        if enabled:
            lflag |= termios.ECHO
        else:
            lflag &= ~termios.ECHO

        termios.tcsetattr(fd, termios.TCSANOW,
                          [iflag, oflag, cflag, lflag, ispeed, ospeed, cc])

    @staticmethod
    def clear():
        sys.stdout.write("\x1b[2J\x1b[H")
        sys.stdout.flush()

    @staticmethod
    def print_error(msg):
        print(Colors.red(msg))

    @staticmethod
    def print_success(msg):
        print(Colors.green(msg))

    @staticmethod
    def get_string(prompt):
        while True:
            value = input(prompt).strip()

            if not value:
                print(Colors.red("Value Required!"))
            else:
                return value

    @staticmethod
    def get_password(prompt="Password: "):
        while True:
            value = getpass.getpass(prompt)

            if not value:
                print(Colors.red("Value Required!"))
            else:
                return value

    @staticmethod
    def get_integer(prompt):
        """Gather user input and convert it to an integer

        Will keep trying till the user enters an interger or until they ^C the
        program.
        """
        while True:
            try:
                return int(input(prompt).strip())
            except ValueError:
                print(Colors.red("Invalid Input!"))


def clear_screen():
    """Clear the terminal
    """
    sys.stdout.write("\x1b[2J\x1b[H")
    sys.stdout.flush()


def iterate_forever(func, *args, **kwargs):
    """Iterate over a finite iterator forever

    When the iterator is exhausted will call the function again to generate a
    new iterator and keep iterating.
    """
    output = func(*args, **kwargs)

    while True:
        try:
            playlistItem = next(output)
            playlistItem.prepare_playback()
            yield playlistItem
        except StopIteration:
            output = func(*args, **kwargs)


class SilentPopen(subprocess.Popen):
    """A Popen varient that dumps it's output and error
    """

    def __init__(self, *args, **kwargs):
        self._dev_null = open(os.devnull, "w")
        kwargs["stdin"] = subprocess.PIPE
        kwargs["stdout"] = subprocess.PIPE
        kwargs["stderr"] = self._dev_null
        super(SilentPopen, self).__init__(*args, **kwargs)

    def __del__(self):
        self._dev_null.close()
        super(SilentPopen, self).__del__()



"""
Pandora API Exceptions

This module contains the API exceptions from the Pandora API. The exception
classes are automatically generated from the API spec below and are added to
this module at first import. Exceptions will always be a sub-class of the base
PandoraException.

The name of the exception class is the message name with space removed and each
word capitalized. If the exception message contains a dash the class name will
not contain the dash or anything that follows it.
"""

__API_EXCEPTIONS__ = {
    0: "Internal Server Error",
    1: "Maintenance Mode",
    2: "Missing API Method",
    3: "Missing Auth Token",
    4: "Missing Partner ID",
    5: "Missing User ID",
    6: "Secure Protocol Required",
    7: "Certificate Required",
    8: "Parameter Type Mismatch",
    9: "Parameter Missing",
    10: "Parameter Value Invalid",
    11: "API Version Not Supported",
    12: "Pandora not available in this country",
    13: "Bad Sync Time",
    14: "Unknown Method Name",
    15: "Wrong Protocol - (http/https)",
    1000: "Read Only Mode",
    1001: "Invalid Auth Token",
    1002: "Invalid Partner Login",
    1003: "Listener Not Authorized - Subscription or Trial Expired",
    1004: "User Not Authorized",
    1005: "Station limit reached",
    1006: "Station does not exist",
    1009: "Device Not Found",
    1010: "Partner Not Authorized",
    1011: "Invalid Username",
    1012: "Invalid Password",
    1023: "Device Model Invalid",
    1039: "Too many requests for a new playlist",
    9999: "Authentication Required",
}


class PandoraException(Exception):
    """Pandora API Exception

    Translates exceptions to user readable info.
    """

    code = None
    message = "Unknown Exception"

    def __init__(self, extended_message=""):
        self.extended_message = extended_message
        super(Exception, self).__init__(self.message)

    @classmethod
    def from_code(cls, code, extended_message):
        exc = __API_EXCEPTIONS__.get(code)

        if not exc:
            exc = PandoraException(extended_message)
            exc.code = code
            return exc
        else:
            return exc(extended_message)

    @staticmethod
    def _format_name(name):
        output = []

        for part in name.split(" "):
            if part == "-":
                break
            else:
                output.append(part.capitalize())

        return "".join(output)

    @staticmethod
    def _export_exceptions(export_to):
        for code, api_message in __API_EXCEPTIONS__.items():
            name = PandoraException._format_name(api_message)

            exception = type(name, (PandoraException,), {
                "code": code,
                "message": api_message,
            })

            export_to[name] = __API_EXCEPTIONS__[code] = exception


PandoraException._export_exceptions(locals())


class InvalidUserLogin(InvalidPartnerLogin):
    """Pydora Internal Login Error

    This is thrown around a user login to disambiguate a login that is invalid
    due to user error vs a login that is invalid due to a partner credential
    error. The Pandora API returns 1002 in both cases.
    """

    message = "Invalid User Login"



"""
Pandora API Client

This is a reasonably complete implementation of the Pandora API. It does not
implement any of the undocumented features and does not implement most of the
account management features as they were deemed not terribly useful.

API Spec from: http://6xq.net/playground/pandora-apidoc/
Keys at: http://6xq.net/playground/pandora-apidoc/json/partners/#partners
"""

from .client import BaseAPIClient, APIClient
from .transport import APITransport, Encryptor, DEFAULT_API_HOST
